function [final_curves] = good_cont(points, b, lambda, outfname)
% GOOD_CONT Main function for detecting good continuation in a pattern of points.
%
% [final_curves] = good_cont(points, b, lambda);
%
% Inputs:
%  - points: a set of 2D points in a [0 1]x[0 1] domain
%  - b: number of nearest neighbors to use for exploration
%  - lambda: ratio of size of local density estimation window and triplet
%  - outfname: preffix of the filename where results will be stored.
% Output
%  - final_curves: cell object containing point indexes of the final curves


% IPOL SUBMISSION "An Unsupervised Algorithm for Detecting Good
% Continuation in Dot Patterns"
%
% Version 0.4, April 2017
%
% Copyright (c) 2017 Jose Lezama <jlezama@fing.edu.uy>
%
% Requires the matlab_bgl toolbox:
% https://www.cs.purdue.edu/homes/dgleich/packages/matlab_bgl/
% It is included in the ZIP file but might require compilation
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.

if ~isdeployed
    addpath lib
end

params.b = b;
params.lambda = lambda;
params.EPSILON = 1;
params.W = 1; % size of domain
params.MIN_PRECISION = params.W/256; % minimum precision tested for a triplet

PLOT = 1;

% check points exist
if size(points,1) < 3 % need at least three for a triplet
    final_curves = {}; 
    hfig = figure();
    if size(points,1)>0
        plot(points(:,1), 1-points(:,2), 'k.', 'MarkerSize', 20)
        axis([0 1 0 1]); axis('square'); axis('off');
        hold on
        plot(0,0,'w'); plot(1,1,'w');
    end
    print(hfig, '-dpng', sprintf('%sresults', outfname))
    print(hfig, '-dpng', sprintf('%soriginal_points', outfname));
    return
end

% normalize points
if max(points(:))>1
    points = points/max(points(:));
end


% plot points
if PLOT
    hfig = figure();
    plot(points(:,1), 1-points(:,2), 'k.', 'MarkerSize', 20)
    axis([0 1 0 1]); axis('square'); axis('off');
    hold on
    plot(0,0,'w'); plot(1,1,'w');
    print(hfig, '-dpng', sprintf('%soriginal_points', outfname));
end


%% build triplets
tstart = tic;
[triplets, D] = build_triplets(points, params);
toc(tstart);

T = size(triplets,1);
N = size(points,1);


%% execute Floyd-Warshall
fprintf('Executing Floyd-Warshal on %i triplets...\n', T);
tstart = tic;
[PathDists, P] = floyd_warshall_all_sp(D);
PathDists(logical(eye(size(PathDists)))) = inf;
fprintf('Finished Floyd-Warshal.\n');
toc(tstart)

%% calculate NFA for obtained candidates
NFAs = compute_paths_NFAs(PathDists, P, triplets, N, b);


% find meaningful curves
z=find(NFAs <params.EPSILON);


%% Masking principle
[final_curves, minNFA] = masking_principle(NFAs, z, P, triplets);

fprintf('Minimal NFA: %2.2e\n',minNFA);


% plot curves
if PLOT
    for i = 1:length(final_curves);
        curve = final_curves{i};
        plot(points(curve,1), 1-points(curve,2), 'r.-', 'MarkerSize', 20, 'LineWidth', 2)
    end
    print(hfig, '-dpng', sprintf('%sresults', outfname))
end

