function [triplets, D] = build_triplets(points, params)
% BUILD_TRIPLETS  returns list of triplets and adjacency matrix
% Input:
%  - points: Nx2 array of 2D points in the domain
%  - params: parameters of the algorithm:
%   - params.b: number of nearest neighbors to explroe
%   - params.lambda: ratio of the local area radius to the distance
%     between the first two points of a triplet
%   - params.MIN_PRECISION: minimum precision of the discrete domain
%   - params.EPSILON: meaningfulness threshold
% Output:
%   - triplets: Tx4 array containing [i, j, k, p], the indices of each 
%     point in the triplet and the probability (error) of a triplet
%   - D: TxT triplet adjacency matrix to serve as input for Floyd-Warshall
  
% IPOL SUBMISSION "An Unsupervised Algorithm for Detecting Good 
% Continuation in Dot Patterns"
%
% Version 0.4, April 2017
%
% Copyright (c) 2017 Jose Lezama <jlezama@fing.edu.uy>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.


N = size(points,1);

fprintf('Building triplets from %i points\n', N);

% manually set the maximum chain length to limit the number of accepted
% triplets. this is just an heuristic for accelerating
%
% Ntests = b * N  * sqrt(N); (b =  number of nearest neighbors)
% NFA = Ntests * max_p^(K-2);

if size(points,1)>200
    Emax = (params.EPSILON/(params.b*N*sqrt(N)))^(1/(round(sqrt(N))-2));
else
    Emax = (params.EPSILON/(params.b*N*sqrt(N)))^(1/(N-2));
end

% compute distances between all points
Dists = pdist2(points,points);
Dists(logical(eye(size(Dists))))=inf;
[~, I] = sort(Dists); % columns are sorted

% points reflection for density estimation
points_mirrorred = mirror_points(points,params.W);

% initialze triplets structure
triplets = ones(N*params.b,5)*-1;
count=1;

% construct triplets based on b neighbors
for i = 1:N
    Ii = I(:,i);
    
    p1 = points(i,:);
    
    for jj = 1:min(N,params.b)
        j = Ii(jj);
        
        p2 = points(j,:);
        
        d12 = pdist2p(p2,p1);
        
        if d12<params.MIN_PRECISION % \Delta in the article
            % too small
            continue
        end
        
        % radius of local window
        R = d12*params.lambda;
        
        % count points in local window with mirrorring
        dists_mirror = pdist2(p2, points_mirrorred);
        N_local_points = length(find(dists_mirror<=R))-2;
        
        % ideal point
        pideal = 2*p2-p1;
        
        % look for candidate points starting from symmetric point
        dists_to_symmetric = pdist2(pideal,points);
        [~, Isym] = sort(dists_to_symmetric);
        
        % remove seed points from symmetric neighbors
        Isym(Isym==i)=[];
        Isym(Isym==j)=[];
        
        for kk = 1:min(2, length(Isym))
            k = Isym(kk);
            if dists_to_symmetric(k)>R
                % outside local window
                continue
            end
            
            % p3 = k
            p3 = points(k,:);
            
            % calculate triplet probability
            p_triplet = calc_triplet_prob(p3,pideal, R,N_local_points, params.MIN_PRECISION);
            
            if p_triplet > Emax
                % this triplet would never constitute a meaningful chain
                continue
            else
                % add triplet to list
                triplets(count,:)=[i j k p_triplet N_local_points];
                count = count+1;
            end
        end
    end
end

% clean array of triplets
triplets((triplets(:,1)==-1),:) = [];

T = size(triplets,1);

% sort triplets
[~, I] = sort(triplets(:,4));
triplets = triplets(I,:);

% build adjacency matrix
D = pdist2(triplets(:,[2 3]),triplets(:,[1 2]));
D = (D==0);

% eliminate triplets without neighbors
% note that it will only detect chains of more than one triplet
triplets((sum(D)==0 & (sum(D,2)==0)'),:)=[];

% recompute adjacency matrix
D = pdist2(triplets(:,[2 3]),triplets(:,[1 2]));
D = (D==0);

% calculate sum of probabilities
P = pdist2(triplets(:,4),-triplets(:,4));
D = D.*P;

D = sparse(double(D));

