function [final_curves, minNFA] = masking_principle(NFAs, z, P, triplets)
% MASKING_PRINCIPLE applies masking principle to chains represented in P with
% NFA values in NFA
%
% [final_curves, minNFA] = masking_principle(NFAs, P, triplets)
%
%  Input
%   - NFAs: TxT matrix of NFAs for every path between triplets
%   - z: index of meaningful NFAs
%   - P: TxT Floyd-Warshall matrix of paths
%   - triplets: Tx4 array of triplets ([i, j, k, p])
%  Output
%   - final_curves: cell object containing point indexes of the final curves
%   - minNFA: minimum NFA

% IPOL SUBMISSION "An Unsupervised Algorithm for Detecting Good
% Continuation in Dot Patterns"
%
% Version 0.4, April 2017
%
% Copyright (c) 2017 Jose Lezama <jlezama@fing.edu.uy>
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
%
% You should have received a copy of the GNU Affero General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.


disp('Applying the masking principle...')
tstart = tic;

% sort by NFA
[I, J] = ind2sub(size(NFAs),z);
[VV, II] = sort(NFAs(z));
C = length(I);
final_curves = cell(0,1);

tic
for x=1:C
    xx = II(x);
    
    i = I(xx);
    j = J(xx);
    
    if toc>3
        fprintf('Checking masking for curve %i of %i\n',x,C)
        tic
    end
    
    fw_path = get_fw_path(i,j,P);
    points_in_path = [triplets(fw_path,1); triplets(fw_path(end),2:3)'];
    
    % check maskings
    masked = 0;
    
    for c=1:length(final_curves);
        existing_curve_points = final_curves{c};
        if  sum(ismember(points_in_path,existing_curve_points))>1
            % allow no more than 1 point of intersection
            masked = 1;
            break
        end
    end
    
    if masked
        continue
    else
        % check that no more than 1 points are repeated
        if length(points_in_path)> length(unique(points_in_path))+1
            continue
        end
        
    end
    
    final_curves{end+1}=points_in_path;
end
toc(tstart);


if ~isempty(VV)
    minNFA = VV(1);
else
    minNFA = inf;
end
